import argparse
import os

def check_positive_int(value):
    try:
        value = int(value)
        if value <= 0:
            raise argparse.ArgumentTypeError("{} is not a positive integer".format(value))
    except ValueError:
        raise Exception("{} is not an integer".format(value))
    return value

def check_positive_float(value):
    try:
        value = float(value)
        if value <= 0.0:
            raise argparse.ArgumentTypeError("{} is not a positive float".format(value))
    except ValueError:
        raise Exception("{} is not an float".format(value))
    return value

class Parser:
    def __init__(self, argv):
        
        parser = argparse.ArgumentParser(prog='virpipe', description='%(prog)s is a command line program for identification of virus sequence from sequencing reads.')
        subparsers = parser.add_subparsers(dest='task', required=True, title='tasks', description='valid tasks')

        common_parser = argparse.ArgumentParser(add_help=False, argument_default=argparse.SUPPRESS)
        common_parser.add_argument('--image', '-i', required=True)
        common_parser.add_argument('--prefix', '-p', required=True)
        common_parser.add_argument('--outdir', '-o')
        common_parser.add_argument('--resume', '-r', action='store_true', default=False)

        platform_parser = argparse.ArgumentParser(add_help=False, argument_default=argparse.SUPPRESS)
        platform_parser.add_argument('--platform', nargs='?', choices=['illumina', 'nanopore'], required=True)

        fastq_input_parser = argparse.ArgumentParser(add_help=False, argument_default=argparse.SUPPRESS)
        fastq_input_parser.add_argument('--fastq', required=True, type=os.path.abspath)
        fastq_input_parser.add_argument('--fastq2', type=os.path.abspath)

        fasta_input_parser = argparse.ArgumentParser(add_help=False, argument_default=argparse.SUPPRESS)
        fasta_input_parser.add_argument('--fasta', required=True, type=os.path.abspath)

        qc_parser = subparsers.add_parser('qc', parents=[common_parser, platform_parser, fastq_input_parser], argument_default=argparse.SUPPRESS)
        qc_parser.add_argument('--nanopore-min-read-quality', nargs='?', type=check_positive_int, default=8)
        qc_parser.add_argument('--nanopore-min-read-length', nargs='?', type=check_positive_int, default=200)

        preprocess_parser = subparsers.add_parser('preprocess', parents=[common_parser, platform_parser, fastq_input_parser], argument_default=argparse.SUPPRESS)
        preprocess_parser.add_argument('--illumina-min-read-quality', nargs='?', type=check_positive_int, default=8)
        preprocess_parser.add_argument('--nanopore-min-read-quality', nargs='?', type=check_positive_int, default=8)
        preprocess_parser.add_argument('--nanopore-min-read-length', nargs='?', type=check_positive_int, default=200)
        
        remove_host_parser = subparsers.add_parser('remove-host', parents=[common_parser, platform_parser, fastq_input_parser], argument_default=argparse.SUPPRESS)
        remove_host_parser.add_argument('--host-genome', nargs='?', required=True, type=os.path.abspath)

        map_parser = subparsers.add_parser('map', parents=[common_parser, platform_parser, fastq_input_parser], argument_default=argparse.SUPPRESS)
        map_parser.add_argument('--save-map-bam', action='store_true', default=False)
        map_parser.add_argument('--ref', nargs='*', type=os.path.abspath)
        map_parser.add_argument('--dir-ref', nargs='?', type=os.path.abspath)
        map_parser.add_argument('--min-avg-cov', nargs='?', type=check_positive_float, default=1.0)

        classify_taxonomy_parser = subparsers.add_parser('classify-taxonomy', parents=[common_parser, platform_parser, fastq_input_parser], argument_default=argparse.SUPPRESS)
        classify_taxonomy_parser.add_argument('--kraken2-confidence-threshold', nargs='?', type=check_positive_float, default=0.1)
        classify_taxonomy_parser.add_argument('--kraken2-db', nargs='?', type=os.path.abspath)
        classify_taxonomy_parser.add_argument('--centrifuge-db', nargs='?', type=os.path.abspath)

        assemble_parser = subparsers.add_parser('assemble', parents=[common_parser, platform_parser, fastq_input_parser], argument_default=argparse.SUPPRESS)
        assemble_parser.add_argument('--assembly-tool', nargs='?', choices=['spades', 'megahit', 'canu', 'flye'])
        assemble_parser.add_argument('--min-contig-length', nargs='?', type=check_positive_int, default=600)

        polish_parser = subparsers.add_parser('polish', parents=[common_parser, fasta_input_parser], argument_default=argparse.SUPPRESS)
        polish_parser.add_argument('--reads', required=True, type=os.path.abspath)

        blast_parser = subparsers.add_parser('blast', parents=[common_parser, fasta_input_parser], argument_default=argparse.SUPPRESS)
        blast_parser.add_argument('--blast-db', nargs='?', required=True, type=os.path.abspath)
        blast_parser.add_argument('--min-evalue', nargs='?', default='1.0e-5')
        blast_parser.add_argument('--min-blast-aln-len', nargs='?', type=check_positive_int, default=100)
        blast_parser.add_argument('--taxonomizr-db', nargs='?', required=True, type=os.path.abspath)

        zoonotic_rank_parser = subparsers.add_parser('zoonotic-rank', parents=[common_parser, fasta_input_parser], argument_default=argparse.SUPPRESS)

        all_parser = subparsers.add_parser('all', parents=[common_parser, platform_parser, fastq_input_parser], argument_default=argparse.SUPPRESS)
        all_parser.add_argument('--save-preprocessed', action='store_true', default=False)
        all_parser.add_argument('--illumina-min-read-quality', nargs='?', type=check_positive_int, default=8)
        all_parser.add_argument('--nanopore-min-read-quality', nargs='?', type=check_positive_int, default=8)
        all_parser.add_argument('--nanopore-min-read-length', nargs='?', type=check_positive_int, default=200)
        all_parser.add_argument('--save-host-removed', action='store_true', default=False)
        all_parser.add_argument('--host-genome', nargs='?', type=os.path.abspath)
        all_parser.add_argument('--save-map-bam', action='store_true', default=False)
        all_parser.add_argument('--ref', nargs='*', type=os.path.abspath)
        all_parser.add_argument('--dir-ref', nargs='?', type=os.path.abspath)
        all_parser.add_argument('--min-avg-cov', nargs='?', type=check_positive_float, default=1.0)
        all_parser.add_argument('--kraken2-confidence-threshold', nargs='?', type=check_positive_float, default=0.1)
        all_parser.add_argument('--kraken2-db', nargs='?', type=os.path.abspath)
        all_parser.add_argument('--centrifuge-db', nargs='?', type=os.path.abspath)
        all_parser.add_argument('--assembly-tool', nargs='?', choices=['spades', 'megahit', 'canu', 'flye'])
        all_parser.add_argument('--min-contig-length', nargs='?', type=check_positive_int, default=600)
        all_parser.add_argument('--reads', type=os.path.abspath)
        all_parser.add_argument('--blast-db', nargs='?', type=os.path.abspath)
        all_parser.add_argument('--min-evalue', nargs='?', default='1.0e-5')
        all_parser.add_argument('--min-blast-aln-len', nargs='?', type=check_positive_int, default=100)
        all_parser.add_argument('--taxonomizr-db', nargs='?', type=os.path.abspath)

        all_parser.add_argument('--skip-qc', action='store_true', default=False)
        all_parser.add_argument('--skip-preprocess', action='store_true', default=False)
        all_parser.add_argument('--skip-remove-host', action='store_true', default=False)
        all_parser.add_argument('--skip-map', action='store_true', default=False)
        all_parser.add_argument('--skip-classify-taxonomy', action='store_true', default=False)
        all_parser.add_argument('--skip-assemble', action='store_true', default=False)
        all_parser.add_argument('--skip-polish', action='store_true', default=False)
        all_parser.add_argument('--skip-blast', action='store_true', default=False)
        all_parser.add_argument('--do-zoonotic-rank', action='store_true', default=False)

        all_parser.add_argument('--args-from-file', nargs='?', type=os.path.abspath)

        self.args = parser.parse_args(argv)

    def get_args(self):
        return vars(self.args) # return as dictionary


import sys
from .arguments.parser import Parser
from .arguments.validation import Validator
from .command.nextflow import CommandConverter
from .command.run_docker import DockerRunner

def main():
    p = Parser(sys.argv[1:])
    
    args = p.get_args()

    args = Validator(args).validate()

    nextflow_command = CommandConverter(args).generate_nextflow_script_command()

    DockerRunner(args).run(nextflow_command)


import os
import sys

def is_gz_file(filepath):
    with open(filepath, 'rb') as test_f:
        return test_f.read(2) == b'\x1f\x8b'

class DockerRunner():
    def __init__(self, args):
        self.args = args
        self.prefix = args.get('prefix')
        self.image = args.get('image')
        self.volumes = {os.getcwd(): {'bind': '/home/user/run', 'mode': 'rw'}}
        self.bind_input()
        self.bind_database()

    def bind_input(self):
        base_input_dir = "/home/user/input"
        for name, path in self.args.items():
            mount_path = ''
            if name == 'ref':
                for i, path in enumerate(self.args['ref']):
                    mount_path = f"{base_input_dir}/{self.prefix}_ref_{i+1}.fasta"
                    self.volumes[path] = {'bind': mount_path, 'mode': 'ro'}
                continue

            if name == 'fastq':
                if is_gz_file(path):
                    mount_path = f"{base_input_dir}/{self.prefix}.fastq.gz"
                else:
                    mount_path = f"{base_input_dir}/{self.prefix}.fastq"
            elif name == 'fastq2':
                if is_gz_file(path):
                    mount_path = f"{base_input_dir}/{self.prefix}_2.fastq.gz"
                else:
                    mount_path = f"{base_input_dir}/{self.prefix}_2.fastq"
            elif name == 'fasta':
                mount_path = f"{base_input_dir}/{self.prefix}.fasta"
            elif name == 'host_genome':
                mount_path = f"{base_input_dir}/{self.prefix}_host_genome.fasta"
            elif name == 'reads':
                if is_gz_file(path):
                    mount_path = f"{base_input_dir}/{self.prefix}_reads.fastq.gz"
                else:
                    mount_path = f"{base_input_dir}/{self.prefix}_reads.fastq"
            elif name == 'dir_ref':
                mount_path = f"{base_input_dir}/dir-ref"

            if mount_path != '':
                self.volumes[path] = {'bind': mount_path, 'mode': 'ro'}

    def bind_database(self):
        base_database_dir = "/home/user/databases"
        for name, path in self.args.items():
            mount_path = ''
            if name == 'kraken2_db':
                mount_path = f"{base_database_dir}/kraken2-db"
            elif name == 'centrifuge_db':
                mount_path = f"{base_database_dir}/centrifuge-db"
            elif name == 'blast_db':
                mount_path = f"{base_database_dir}/blast-db"
            elif name == 'taxonomizr_db':
                mount_path = f"{base_database_dir}/taxonomizr-db.sql"

            if mount_path != '':
                if name in ['centrifuge_db', 'blast_db']:
                    head, _ = os.path.split(path)
                    self.volumes[head] = {'bind': mount_path, 'mode': 'ro'}
                else:
                    self.volumes[path] = {'bind': mount_path, 'mode': 'ro'}

    def run(self, nextflow_command):
        client = docker.from_env()

        container = client.containers.run(image=self.image,
                              command=nextflow_command,
                              detach=True,
                              volumes=self.volumes,
                              working_dir='/home/user/run')
        
        out = container.logs(stdout=True, stderr=False, stream=True)
        err = container.logs(stdout=False, stderr=True, stream=True)

        for line in out:
            if line.decode().strip():
                print(line.decode().strip())
            
        for line in err:
            if line.decode().strip():
                print(line.decode().strip(), file=sys.stderr)
            
        container.remove()


from setuptools import setup, find_packages

setup_requires = [
    ]

install_requires = [
    'docker',
    'pyfastx',
    ]

setup(
    name='VirPipe',
    version='1.0.5',
    description='VirPipe is a easy-to-use computational pipeline to identify virus sequences from high-throughput sequencing reads.',
    author='Kijin Kim',
    author_email='skkujin@gmail.com',
    url='https://github.com/KijinKims/VirPipe',
    packages=find_packages(),
    install_requires=install_requires,
    setup_requires=setup_requires,
    license='MIT',
    entry_points={
        'console_scripts': [
            'virpipe = virpipe.__main__:main',
            ],
        },
    )



import os
import pyfastx
import warnings

class InputError(Exception):
    def __init__(self, message):
        self.message = message
        
def is_gz_file(filepath):
    with open(filepath, 'rb') as test_f:
        return test_f.read(2) == b'\x1f\x8b'

class Validator():
    def __init__(self, args):

        self.args = args
        self.task = args.get('task')

    def validate(self):
        self.validate_common_params()
        self.validate_input_file()

        if self.task == 'all':
            self.validate_all()
        elif self.task == 'qc':
            self.validate_qc()
        elif self.task == 'preprocess':
            self.validate_preprocess()
        elif self.task == 'remove-host':
            self.validate_remove_host()
        elif self.task == 'classify-taxonomy':
            self.validate_classify_taxonomy()
        elif self.task == 'assemble':
            self.validate_assemble()
        elif self.task == 'polish':
            self.validate_polish()
        elif self.task == 'map':
            self.validate_map()
        elif self.task == 'blast':
            self.validate_blast()
        elif self.task == 'zoonotic-rank':
            self.validate_zoonotic_rank()

        return self.args

    def validate_common_params(self):

        # --outdir copied from --prefix, if not given
        if not self.args.get('outdir'):
            self.args['outdir'] = self.args.get('prefix')

        # --config file check
        if self.args.get('config'):
            self.file_exists(self.args.get('config'))

    def validate_input_file(self):

        platform = self.args.get('platform')

        # if --platform given
        if platform:
            # nanopore
            if platform == 'nanopore':
                self.is_fastq(self.args.get('fastq'))
                if self.args.get('fastq2'):
                    raise InputError("--fastq2 is not required for nanopore platform.")

            # illumina
            elif platform == 'illumina':
                self.is_fastq(self.args.get('fastq'))
                if not self.args.get('fastq2'):
                    raise InputError("Only paired input is allowed for illumina platform. --fastq2 is required.")
                self.is_fastq(self.args.get('fastq2'))
        else:
            self.is_fasta(self.args.get('fasta'))

    def file_exists(self, filename):
        if os.path.exists(filename):
            return True
        else:
            raise FileNotFoundError("The file %s does not exists." % filename)

    def dir_exists(self, dirname):
        if os.path.exists(dirname):
            return True
        else:
            raise FileNotFoundError("The directory %s does not exists." % dirname)

    def is_fasta(self, filename):
        pyfastx.Fasta(filename, build_index=False)
        return True

    def is_fastq(self, filename):
        pyfastx.Fastq(filename, build_index=False)
        return True

    def validate_qc(self):
        pass

    def validate_preprocess(self):
        pass

    def validate_remove_host(self):
        if not self.args.get('host_genome'):
            raise InputError("--host-genome is required.  If you don't want to remove host genome, use --skip-remove-host.")
        self.is_fasta(self.args.get('host_genome'))

    def validate_classify_taxonomy(self):
        platform = self.args.get('platform')

        if platform == 'nanopore':
            if self.args.get('centrifuge_db'):
                self.dir_exists(os.path.dirname(self.args.get('centrifuge_db')))
            else:
                raise InputError("--centrifuge-db is required for nanopore platform. If you don't want to classify taxonomy, use --skip-classify-taxonomy.")

        elif platform == 'illumina':
            if self.args.get('kraken2_db'):
                self.dir_exists(self.args.get('kraken2_db'))
            else:
                raise InputError("--kraken2-db is required for illumina platform. If you don't want to classify taxonomy, use --skip-classify-taxonomy.")
                

    def validate_assemble(self):
        platform = platform = self.args.get('platform')

        if self.args.get('assembly_tool'):
            if platform == 'nanopore' and self.args.get('assembly_tool') == 'spades':
                raise InputError("--assembly-tool can be chosen from ['megahit', 'canu', 'flye'] for nanopore platform")
            elif platform == 'illumina' and self.args.get('assembly_tool') != 'spades':
                raise InputError("--assembly-tool can be chosen from ['spades'] for illumina platform")
            
        else:
            platform = self.args.get('platform')

            if platform == 'nanopore':
                warnings.warn("--assembly-tool is not given. Flye will be used as a default.")
                self.args['assembly_tool'] = 'flye'
            else:
                warnings.warn("--assembly-tool is not given. Spades will be used as a default.")
                self.args['assembly_tool'] = 'spades'

    def validate_polish(self):
        if not self.args.get('reads'):
            raise InputError("--reads is required. If you don't want to polish, use --skip-polish.")
        self.is_fastq(self.args.get('reads'))

    def validate_map(self):
        if self.args.get('ref'):
            for ref in self.args.get('ref'):
                self.is_fasta(ref)
        if self.args.get('dir_ref'):
            self.dir_exists(self.args.get('dir_ref'))

    def validate_blast(self):
        if not self.args.get('blast_db'):
            raise InputError("--blast-db is required. If you don't want to run blast, use --skip-blast.")
        self.dir_exists(os.path.dirname(self.args.get('blast_db')))

        if not self.args.get('taxonomizr_db'):
            raise InputError("--taxonomizr-db is required. If you don't want to run blast, use --skip-blast.")
        self.file_exists(self.args.get('taxonomizr_db'))
        if is_gz_file(self.args.get('taxonomizr_db')):
            raise InputError("taxonomizr db is in gzipped format. It should be decompressed.")

    def validate_zoonotic_rank(self):
        pass

    def parse_args_from_file(self, filename):
        int_k = ['illumina_min_read_quality',
                 'illumina_min_read_length',
                 'nanopore_min_read_quality',
                 'min_contig_length',
                 'min_blast_aln_len']
        
        float_k = ['min_avg_dep',
                   'kraken2_confidence_threshold']

        with open(filename) as f_in:
            lines = filter(None, (line.rstrip() for line in f_in)) # remove empty line
            lines = filter(line.startswith('#') for line in lines) # remove comment line
            for line in lines:
                arg_key, arg_value = line.split('=')
                k = arg_key.strip()
                v = arg_value.strip()

                if k.startswith('save') or k.startswith('skip') or k.startswith('do') or k =='resume':
                    self.args[k] = bool(v)
                elif k in int_k:
                    self.args[k] = int(v)
                elif k in float_k:
                    self.args[k] = float(v)
                else:
                    self.args[k] = k

    def validate_all(self):
        if self.args.get('args_from_file'):
            self.file_exists(self.args.get('args_from_file'))
            self.parse_args_from_file(self.args.get('args_from_file'))

        # no polish step for illumina input
        if self.args.get('platform') == 'illumina':
            self.args['skip_polish'] = True

        if not self.args.get('skip_qc'):
            self.validate_qc()
        if not self.args.get('skip_preprocess'):
            self.validate_preprocess()
        if not self.args.get('skip_remove_host'):
            self.validate_remove_host()
        if not self.args.get('skip_classify_taxonomy'):
            self.validate_classify_taxonomy()
        if not self.args.get('skip_assemble'):
            self.validate_assemble()
        if not self.args.get('skip_polish'):
            self.validate_polish()
        if not self.args.get('skip_map'):
            self.validate_map()
        if not self.args.get('skip_blast'):
            self.validate_blast()
        if self.args.get('do_zoonotic_rank'):
            self.validate_zoonotic_rank()


import os

def is_gz_file(filepath):
    with open(filepath, 'rb') as test_f:
        return test_f.read(2) == b'\x1f\x8b'

class CommandConverter():
        def __init__(self, args):

            self.comp_list = []
            
            self.args = args
            self.prefix = self.args['prefix']
            base_script_dir = "/home/user/nextflow_scripts"
            self.comp_list.append("nextflow run")
            script = f"{base_script_dir}/{self.args['task']}.nf"
            config = f"-c {base_script_dir}/{self.args['task']}.config"
            self.comp_list.append(script)
            self.comp_list.append(config)

            for k, v in args.items():
                if k == 'image':
                    continue
                elif k.startswith('save') or k.startswith('skip') or k.startswith('do'):
                     if v:
                        self.comp_list.append(f"--{k}")
                elif k == 'resume':
                    if v:
                        self.comp_list.append(f"-{k}")
                elif 'db' in k:
                    if k in ['centrifuge_db', 'blast_db']:
                        self.comp_list.append(self.bind_database(k, v))
                    else:
                        self.comp_list.append(self.bind_database(k, v))
                elif k in ['fastq', 'fastq2', 'fasta', 'host_genome', 'reads', 'ref', 'dir-ref']:
                    self.comp_list.append(self.bind_input(k, v))
                else:
                    self.comp_list.append(f"--{k} {v}")

        def bind_database(self, name, path):
            base_database_dir = "/home/user/databases"
            if name == 'kraken2_db':
                return f"--{name} {base_database_dir}/kraken2-db"
            elif name == 'centrifuge_db':
                _, tail = os.path.split(path)
                return f"--centrifuge_db_dir {base_database_dir}/centrifuge-db --centrifuge_db_name {tail}"
            elif name == 'blast_db':
                _, tail = os.path.split(path)
                return f"--blast_db_dir {base_database_dir}/blast-db --blast_db_name {tail}"
            elif name == 'taxonomizr_db':
                return f"--{name} {base_database_dir}/taxonomizr-db.sql"

        def bind_input(self, name, path):
            base_input_dir = "/home/user/input"
            if name == 'fastq':
                if is_gz_file(path):
                    return f"--{name} {base_input_dir}/{self.prefix}.fastq.gz"
                else:
                    return f"--{name} {base_input_dir}/{self.prefix}.fastq"
            elif name == 'fastq2':
                if is_gz_file(path):
                    return f"--{name} {base_input_dir}/{self.prefix}_2.fastq.gz"
                else:
                    return f"--{name} {base_input_dir}/{self.prefix}_2.fastq"
            elif name == 'fasta':
                return f"--{name} {base_input_dir}/{self.prefix}.fasta"
            elif name == 'host_genome':
                return f"--{name} {base_input_dir}/{self.prefix}_host_genome.fasta"
            elif name == 'reads':
                if is_gz_file(path):
                    return f"--{name} {base_input_dir}/{self.prefix}_reads.fastq.gz"
                else:
                    return f"--{name} {base_input_dir}/{self.prefix}_reads.fastq.gz"
            elif name == 'ref':
                return f"--{name} " + ' '.join([f"{base_input_dir}/{self.prefix}_ref_{i+1}.fasta" for i in range(len(self.args['ref']))])
            elif name == 'dir_ref':
                return f"--{name} {base_input_dir}/dir-ref"

        def generate_nextflow_script_command(self):
            return ' '.join(self.comp_list)